@@ -7,24 +7,23 @@ import os
 import sys
 import weakref
 from collections import deque
-from enum import Enum
+from enum import IntEnum
+from itertools import chain
 
-from .child import cwd_of_process
 from .config import build_ansi_color_table
 from .constants import (
     ScreenGeometry, WindowGeometry, appname, get_boss, wakeup
 )
 from .fast_data_types import (
     BLIT_PROGRAM, CELL_BG_PROGRAM, CELL_FG_PROGRAM, CELL_PROGRAM,
-    CELL_SPECIAL_PROGRAM, CSI, CURSOR_PROGRAM, DCS, DECORATION, DIM,
-    GRAPHICS_PREMULT_PROGRAM, GRAPHICS_PROGRAM, OSC, REVERSE, SCROLL_FULL,
-    SCROLL_LINE, SCROLL_PAGE, STRIKETHROUGH, Screen, add_window,
-    cell_size_for_window, compile_program, get_clipboard_string,
-    glfw_post_empty_event, init_cell_program, init_cursor_program,
+    CELL_SPECIAL_PROGRAM, CSI, DCS, DECORATION, DIM, GRAPHICS_PREMULT_PROGRAM,
+    GRAPHICS_PROGRAM, OSC, REVERSE, SCROLL_FULL, SCROLL_LINE, SCROLL_PAGE,
+    STRIKETHROUGH, Screen, add_window, cell_size_for_window, compile_program,
+    get_clipboard_string, glfw_post_empty_event, init_cell_program,
     set_clipboard_string, set_titlebar_color, set_window_render_data,
     update_window_title, update_window_visibility, viewport_for_window
 )
-from .keys import keyboard_mode_name
+from .keys import keyboard_mode_name, extended_key_event, defines
 from .rgb import to_color
 from .terminfo import get_capabilities
 from .utils import (
@@ -33,7 +32,7 @@ from .utils import (
 )
 
 
-class DynamicColor(Enum):
+class DynamicColor(IntEnum):
     default_fg, default_bg, cursor_color, highlight_fg, highlight_bg = range(1, 6)
 
 
@@ -59,6 +58,10 @@ def calculate_gl_geometry(window_geometry, viewport_width, viewport_height, cell
 def load_shader_programs(semi_transparent=0):
     compile_program(BLIT_PROGRAM, *load_shaders('blit'))
     v, f = load_shaders('cell')
+
+    def color_as_vec3(x):
+        return 'vec3({}, {}, {})'.format(x.red / 255, x.green / 255, x.blue / 255)
+
     for which, p in {
             'SIMPLE': CELL_PROGRAM,
             'BACKGROUND': CELL_BG_PROGRAM,
@@ -66,7 +69,12 @@ def load_shader_programs(semi_transparent=0):
             'FOREGROUND': CELL_FG_PROGRAM,
     }.items():
         vv, ff = v.replace('WHICH_PROGRAM', which), f.replace('WHICH_PROGRAM', which)
-        for gln, pyn in {'REVERSE_SHIFT': REVERSE, 'STRIKE_SHIFT': STRIKETHROUGH, 'DIM_SHIFT': DIM, 'DECORATION_SHIFT': DECORATION}.items():
+        for gln, pyn in {
+                'REVERSE_SHIFT': REVERSE,
+                'STRIKE_SHIFT': STRIKETHROUGH,
+                'DIM_SHIFT': DIM,
+                'DECORATION_SHIFT': DECORATION,
+        }.items():
             vv = vv.replace('{{{}}}'.format(gln), str(pyn), 1)
         if semi_transparent:
             vv = vv.replace('#define NOT_TRANSPARENT', '#define TRANSPARENT')
@@ -80,33 +88,58 @@ def load_shader_programs(semi_transparent=0):
         ff = f.replace('ALPHA_TYPE', which)
         compile_program(p, v, ff)
     init_cell_program()
-    compile_program(CURSOR_PROGRAM, *load_shaders('cursor'))
-    init_cursor_program()
 
 
 def setup_colors(screen, opts):
     screen.color_profile.update_ansi_color_table(build_ansi_color_table(opts))
+    cursor_text_color = opts.cursor_text_color or (12, 12, 12)
+    cursor_text_color_as_bg = 3 if cursor_text_color is None else 1
     screen.color_profile.set_configured_colors(*map(color_as_int, (
-        opts.foreground, opts.background, opts.cursor, opts.selection_foreground, opts.selection_background)))
+        opts.foreground, opts.background, opts.cursor,
+        cursor_text_color, (0, 0, cursor_text_color_as_bg),
+        opts.selection_foreground, opts.selection_background)
+    ))
+
+
+def text_sanitizer(as_ansi, add_wrap_markers):
+    import re
+    pat = re.compile(r'\033\[.+?m')
+
+    def remove_wrap_markers(line):
+        return line.replace('\r', '')
+
+    def remove_sgr(line):
+        return pat.sub('', line)
+
+    def remove_both(line):
+        return pat.sub('', line.replace('\r', ''))
+
+    if as_ansi:
+        return remove_both if add_wrap_markers else remove_sgr
+    return remove_wrap_markers
 
 
 class Window:
 
     def __init__(self, tab, child, opts, args, override_title=None):
         self.action_on_close = None
+        self.layout_data = None
+        self.pty_resized_once = False
         self.needs_attention = False
         self.override_title = override_title
         self.overlay_window_id = None
         self.overlay_for = None
         self.default_title = os.path.basename(child.argv[0] or appname)
         self.child_title = self.default_title
+        self.title_stack = deque(maxlen=10)
+        self.allow_remote_control = child.allow_remote_control
         self.id = add_window(tab.os_window_id, tab.id, self.title)
-        self.clipboard_control_buffers = {'p': '', 'c': ''}
         if not self.id:
             raise Exception('No tab with id: {} in OS Window: {} was found, or the window counter wrapped'.format(tab.id, tab.os_window_id))
         self.tab_id = tab.id
         self.os_window_id = tab.os_window_id
         self.tabref = weakref.ref(tab)
+        self.clipboard_control_buffers = {'p': '', 'c': ''}
         self.destroyed = False
         self.click_queue = deque(maxlen=3)
         self.geometry = WindowGeometry(0, 0, 0, 0, 0, 0)
@@ -125,12 +158,16 @@ class Window:
         return 'Window(title={}, id={}, overlay_for={}, overlay_window_id={})'.format(
                 self.title, self.id, self.overlay_for, self.overlay_window_id)
 
-    def as_dict(self):
+    def as_dict(self, is_focused=False):
         return dict(
             id=self.id,
+            is_focused=is_focused,
             title=self.override_title or self.title,
             pid=self.child.pid,
-            cwd=self.child.current_cwd or self.child.cwd, cmdline=self.child.cmdline
+            cwd=self.child.current_cwd or self.child.cwd,
+            cmdline=self.child.cmdline,
+            env=self.child.environ,
+            foreground_processes=self.child.foreground_processes
         )
 
     @property
@@ -151,6 +188,12 @@ class Window:
                 if pat.search(x) is not None:
                     return True
             return False
+        if field == 'env':
+            key_pat, val_pat = pat
+            for key, val in self.child.environ.items():
+                if key_pat.search(key) is not None and (
+                        val_pat is None or val_pat.search(val) is not None):
+                    return True
         return False
 
     def set_visible_in_layout(self, window_idx, val):
@@ -182,6 +225,9 @@ class Window:
             sg = self.update_position(new_geometry)
             self.needs_layout = False
             boss.child_monitor.resize_pty(self.id, *current_pty_size)
+            if not self.pty_resized_once:
+                self.pty_resized_once = True
+                self.child.mark_terminal_ready()
         else:
             sg = self.update_position(new_geometry)
         self.geometry = g = new_geometry
@@ -224,7 +270,7 @@ class Window:
 
     # screen callbacks {{{
     def use_utf8(self, on):
-        get_boss().child_monitor.set_iutf8(self.window_id, on)
+        get_boss().child_monitor.set_iutf8(self.id, on)
 
     def focus_changed(self, focused):
         if focused:
@@ -352,6 +398,8 @@ class Window:
 
     def clipboard_control(self, data):
         where, text = data.partition(';')[::2]
+        if not where:
+            where = 's0'
         if text == '?':
             response = None
             if 's' in where or 'c' in where:
@@ -390,6 +438,16 @@ class Window:
                         write('c', set_clipboard_string)
                 if 'write-primary' in self.opts.clipboard_control:
                     write('p', set_primary_selection)
+
+    def manipulate_title_stack(self, pop, title, icon):
+        if title:
+            if pop:
+                if self.title_stack:
+                    self.child_title = self.title_stack.pop()
+                    self.title_updated()
+            else:
+                if self.child_title:
+                    self.title_stack.append(self.child_title)
     # }}}
 
     def text_for_selection(self):
@@ -402,34 +460,48 @@ class Window:
             self.screen.reset_callbacks()
         self.screen = None
 
-    def as_text(self, as_ansi=False, add_history=False, add_wrap_markers=False):
+    def as_text(self, as_ansi=False, add_history=False, add_wrap_markers=False, alternate_screen=False):
         lines = []
-        add_history = add_history and not self.screen.is_using_alternate_linebuf()
-        f = self.screen.as_text_non_visual if add_history else self.screen.as_text
+        add_history = add_history and not (self.screen.is_using_alternate_linebuf() ^ alternate_screen)
+        if alternate_screen:
+            f = self.screen.as_text_alternate
+        else:
+            f = self.screen.as_text_non_visual if add_history else self.screen.as_text
         f(lines.append, as_ansi, add_wrap_markers)
         if add_history:
             h = []
+            self.screen.historybuf.pagerhist_as_text(h.append)
+            if not as_ansi or not add_wrap_markers:
+                sanitizer = text_sanitizer(as_ansi, add_wrap_markers)
+                h = list(map(sanitizer, h))
             self.screen.historybuf.as_text(h.append, as_ansi, add_wrap_markers)
-            lines = h + lines
+            lines = chain(h, lines)
         return ''.join(lines)
 
     @property
     def cwd_of_child(self):
-        # TODO: Maybe use the cwd of the leader of the foreground process
-        # group in the session of the child process?
-        pid = self.child.pid
-        if pid is not None:
-            return cwd_of_process(pid) or None
+        return self.child.foreground_cwd or self.child.current_cwd
+
+    def pipe_data(self, text, has_wrap_markers=False):
+        text = text or ''
+        if has_wrap_markers:
+            text = text.replace('\r\n', '\n').replace('\r', '\n')
+        lines = text.count('\n')
+        input_line_number = (lines - (self.screen.lines - 1) - self.screen.scrolled_by)
+        return {
+            'input_line_number': input_line_number, 'scrolled_by': self.screen.scrolled_by,
+            'cursor_x': self.screen.cursor.x + 1, 'cursor_y': self.screen.cursor.y + 1,
+            'lines': self.screen.lines, 'columns': self.screen.columns,
+            'text': text
+        }
 
     # actions {{{
 
     def show_scrollback(self):
-        data = self.as_text(as_ansi=True, add_history=True, add_wrap_markers=True)
-        data = data.replace('\r\n', '\n').replace('\r', '\n')
-        lines = data.count('\n')
-        input_line_number = (lines - (self.screen.lines - 1) - self.screen.scrolled_by)
-        cmd = [x.replace('INPUT_LINE_NUMBER', str(input_line_number)) for x in self.opts.scrollback_pager]
-        get_boss().display_scrollback(self, data, cmd)
+        text = self.as_text(as_ansi=True, add_history=True, add_wrap_markers=True)
+        data = self.pipe_data(text, has_wrap_markers=True)
+        cmd = [x.replace('INPUT_LINE_NUMBER', str(data['input_line_number'])) for x in self.opts.scrollback_pager]
+        get_boss().display_scrollback(self, data['text'], cmd)
 
     def paste(self, text):
         if text and not self.destroyed:
@@ -441,6 +513,10 @@ class Window:
                     if len(text) == len(new_text):
                         break
                     text = new_text
+            else:
+                # Workaround for broken editors like nano that cannot handle
+                # newlines in pasted text see https://github.com/kovidgoyal/kitty/issues/994
+                text = b'\r'.join(text.splitlines())
             self.screen.paste(text)
 
     def copy_to_clipboard(self):
@@ -448,6 +524,15 @@ class Window:
         if text:
             set_clipboard_string(text)
 
+    def copy_or_interrupt(self):
+        text = self.text_for_selection()
+        if text:
+            set_clipboard_string(text)
+        else:
+            mode = keyboard_mode_name(self.screen)
+            text = extended_key_event(defines.GLFW_KEY_C, defines.GLFW_MOD_CONTROL, defines.GLFW_PRESS) if mode == 'kitty' else b'\x03'
+            self.write_to_child(text)
+
     def pass_selection_to_program(self, *args):
         cwd = self.cwd_of_child
         text = self.text_for_selection()
